import time
from os import makedirs
from os.path import abspath, exists, join

import matplotlib.pyplot as plt
from matplotlib.patches import Polygon

from extremitypathfinder.extremitypathfinder import PolygonEnvironment

EXPORT_RESOLUTION = 200  # dpi
EXPORT_SIZE_X = 19.0  # inch
EXPORT_SIZE_Y = 11.0  # inch

POLYGON_SETTINGS = {
    'edgecolor': 'black',
    'fill': False,
    'linewidth': 1.0,
}

SHOW_PLOTS = False
# TODO avoid global variable
PLOTTING_DIR = 'all_plots'


def get_plot_name(file_name='plot'):
    return abspath(join(PLOTTING_DIR, file_name + '_' + str(time.time())[:-7] + '.png'))


def export_plot(fig, file_name):
    fig.set_size_inches(EXPORT_SIZE_X, EXPORT_SIZE_Y, forward=True)
    plt.savefig(get_plot_name(file_name), dpi=EXPORT_RESOLUTION)
    plt.close()


def mark_points(vertex_iter, **kwargs):
    try:
        coordinates = [v.coordinates.tolist() for v in vertex_iter]
    except AttributeError:
        coordinates = [v for v in vertex_iter]
    coords_zipped = list(zip(*coordinates))
    if coords_zipped:  # there might be no vertices at all
        plt.scatter(*coords_zipped, **kwargs)


def draw_edge(v1, v2, c, alpha, **kwargs):
    if type(v1) == tuple:
        x1, y1 = v1
        x2, y2 = v2
    else:
        x1, y1 = v1.coordinates
        x2, y2 = v2.coordinates
    plt.plot([x1, x2], [y1, y2], color=c, alpha=alpha, **kwargs)


def draw_polygon(ax, coords, **kwargs):
    kwargs.update(POLYGON_SETTINGS)
    polygon = Polygon(coords, **kwargs)
    ax.add_patch(polygon)


def draw_boundaries(map, ax):
    # TODO outside light grey
    # TODO fill holes light grey
    draw_polygon(ax, map.boundary_polygon.coordinates)
    for h in map.holes:
        draw_polygon(ax, h.coordinates, facecolor='grey', fill=True)

    mark_points(map.all_vertices, c='black', s=15)
    mark_points(map.all_extremities, c='red', s=50)


def draw_internal_graph(map, ax):
    for start, all_goals in map.graph.get_neighbours():
        for goal in all_goals:
            draw_edge(start, goal, c='red', alpha=0.2, linewidth=2)


def set_limits(map, ax):
    ax.set_xlim((min(map.boundary_polygon.coordinates[:, 0]) - 1, max(map.boundary_polygon.coordinates[:, 0]) + 1))
    ax.set_ylim((min(map.boundary_polygon.coordinates[:, 1]) - 1, max(map.boundary_polygon.coordinates[:, 1]) + 1))


def draw_path(vertex_path):
    # start, path and goal in green
    if vertex_path:
        mark_points(vertex_path, c='g', alpha=0.9, s=50)
        mark_points([vertex_path[0], vertex_path[-1]], c='g', s=100)
        v1 = vertex_path[0]
        for v2 in vertex_path[1:]:
            draw_edge(v1, v2, c='g', alpha=1.0)
            v1 = v2


def draw_loaded_map(map):
    fig, ax = plt.subplots()

    draw_boundaries(map, ax)
    set_limits(map, ax)
    export_plot(fig, 'map_plot')
    if SHOW_PLOTS:
        plt.show()


def draw_prepared_map(map):
    fig, ax = plt.subplots()

    draw_boundaries(map, ax)
    draw_internal_graph(map, ax)
    set_limits(map, ax)
    export_plot(fig, 'prepared_map_plot')
    if SHOW_PLOTS:
        plt.show()


def draw_with_path(map, temp_graph, vertex_path):
    fig, ax = plt.subplots()

    start, goal = vertex_path[0], vertex_path[-1]
    draw_boundaries(map, ax)
    draw_internal_graph(map, ax)
    set_limits(map, ax)

    # additionally draw:
    # new edges yellow
    if start in temp_graph.get_all_nodes():
        for n2, d in temp_graph.edges_from(start):
            draw_edge(start, n2, c='y', alpha=0.7)

    all_nodes = temp_graph.get_all_nodes()
    if goal in all_nodes:
        # edges only run towards goal
        for n1 in all_nodes:
            if goal in temp_graph.get_neighbours_of(n1):
                draw_edge(n1, goal, c='y', alpha=0.7)

    # start, path and goal in green
    draw_path(vertex_path)

    export_plot(fig, 'graph_path_plot')
    if SHOW_PLOTS:
        plt.show()


def draw_only_path(map, vertex_path):
    fig, ax = plt.subplots()

    draw_boundaries(map, ax)
    set_limits(map, ax)
    draw_path(vertex_path)

    export_plot(fig, 'path_plot')
    if SHOW_PLOTS:
        plt.show()


def draw_graph(map, graph):
    fig, ax = plt.subplots()

    all_nodes = graph.get_all_nodes()
    mark_points(all_nodes, c='black', s=30)

    for n in all_nodes:
        x, y = n.coordinates
        neighbours = graph.get_neighbours_of(n)
        for n2 in neighbours:
            x2, y2 = n2.coordinates
            dx, dy = x2 - x, y2 - y
            plt.arrow(x, y, dx, dy, head_width=0.15, head_length=0.5, head_starts_at_zero=False, shape='full',
                      length_includes_head=True)

    set_limits(map, ax)

    export_plot(fig, 'graph_plot')
    if SHOW_PLOTS:
        plt.show()


class PlottingEnvironment(PolygonEnvironment):

    def __init__(self, plotting_dir=PLOTTING_DIR):
        super().__init__()
        global PLOTTING_DIR
        PLOTTING_DIR = plotting_dir
        if not exists(plotting_dir):
            makedirs(plotting_dir)

    def store(self, *args, **kwargs):
        super().store(*args, **kwargs)
        draw_loaded_map(self)

    def prepare(self):
        super().prepare()
        draw_prepared_map(self)

    def find_shortest_path(self, *args, **kwargs):
        # important to not delete the temp graph! for plotting
        vertex_path, distance = super().find_shortest_path(*args, free_space_after=False, **kwargs)

        if self.temp_graph:  # in some cases (e.g. direct path possible) no graph is being created!
            draw_graph(self, self.temp_graph)
            draw_with_path(self, self.temp_graph, vertex_path)
            draw_only_path(self, vertex_path)
            del self.temp_graph  # free the memory

        # extract the coordinates from the path
        return vertex_path, distance



import unittest

import numpy as np
import pytest

from extremitypathfinder.helper_classes import AngleRepresentation
from helpers import proto_test_case


class HelperClassesTest(unittest.TestCase):

    def test_angle_repr(self):
        with pytest.raises(ValueError):
            AngleRepresentation(np.array([0.0, 0.0]))

        #
        # def quadrant_test_fct(input):
        #     np_2D_coord_vector = np.array(input)
        #     return AngleRepresentation(np_2D_coord_vector).quadrant
        #
        # data = [
        #     ([1.0, 0.0], 0.0),
        #     ([0.0, 1.0], 0.0),
        #     ([-1.0, 0.0], 1.0),
        #     ([0.0, -1.0], 3.0),
        #
        #     ([2.0, 0.0], 0.0),
        #     ([0.0, 2.0], 0.0),
        #     ([-2.0, 0.0], 1.0),
        #     ([0.0, -2.0], 3.0),
        #
        #     ([1.0, 1.0], 0.0),
        #     ([-1.0, 1.0], 1.0),
        #     ([-1.0, -1.0], 2.0),
        #     ([1.0, -1.0], 3.0),
        #
        #     ([1.0, 0.00001], 0.0),
        #     ([0.00001, 1.0], 0.0),
        #     ([-1.0, 0.00001], 1.0),
        #     ([0.00001, -1.0], 3.0),
        #
        #     ([1.0, -0.00001], 3.0),
        #     ([-0.00001, 1.0], 1.0),
        #     ([-1.0, -0.00001], 2.0),
        #     ([-0.00001, -1.0], 2.0),
        # ]
        #
        # proto_test_case(data, quadrant_test_fct)

        # TODO test:
        # randomized
        # every quadrant contains angle measures from 0.0 to 1.0
        # angle %360!
        #     rep(p1) > rep(p2) <=> angle(p1) > angle(p2)
        #     rep(p1) = rep(p2) <=> angle(p1) = angle(p2)
        # repr value in [0.0 : 4.0[

        def value_test_fct(input):
            np_2D_coord_vector = np.array(input)
            return AngleRepresentation(np_2D_coord_vector).value

        data = [
            ([1.0, 0.0], 0.0),
            ([0.0, 1.0], 1.0),
            ([-1.0, 0.0], 2.0),
            ([0.0, -1.0], 3.0),

            ([2.0, 0.0], 0.0),
            ([0.0, 2.0], 1.0),
            ([-2.0, 0.0], 2.0),
            ([0.0, -2.0], 3.0),
        ]

        proto_test_case(data, value_test_fct)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(HelperClassesTest)
    unittest.TextTestRunner(verbosity=2).run(suite)
    # unittest.main()



import unittest
from math import sqrt

import pytest

from extremitypathfinder.extremitypathfinder import PolygonEnvironment
from extremitypathfinder.plotting import PlottingEnvironment

# TODO
# PLOT_TEST_RESULTS = True
PLOT_TEST_RESULTS = False
TEST_PLOT_OUTPUT_FOLDER = 'plots'

if PLOT_TEST_RESULTS:
    print('plotting test environment enabled.')
    ENVIRONMENT_CLASS = PlottingEnvironment
    CONSTRUCTION_KWARGS = {"plotting_dir": TEST_PLOT_OUTPUT_FOLDER}
else:
    ENVIRONMENT_CLASS = PolygonEnvironment
    CONSTRUCTION_KWARGS = {}

# size_x, size_y, obstacle_iter
GRID_ENV_PARAMS = (19, 10, [
    # (x,y),
    # obstacles changing boundary
    (0, 1),
    (1, 1),
    (2, 1),
    (3, 1),

    (17, 9),
    (17, 8),
    (17, 7),

    (17, 5),
    (17, 4),
    (17, 3),
    (17, 2),
    (17, 1),
    (17, 0),

    # hole 1
    (5, 5),
    (5, 6),
    (6, 6),
    (6, 7),
    (7, 7),

    # hole 2
    (7, 5),
])

INVALID_DESTINATION_DATA = [
    # outside of map region
    ((-1, 5.0), (17, 0.5)),
    ((17, 0.5), (-1, 5.0)),
    ((20, 5.0), (17, 0.5)),
    ((17, 0.5), (20, 5.0)),
    ((1, -5.0), (17, 0.5)),
    ((17, 0.5), (1, -5.0)),
    ((1, 11.0), (17, 0.5)),
    ((17, 0.5), (1, 11.0)),

    # outside boundary polygon
    ((17.5, 5.0), (17, 0.5)),
    ((17, 0.5), (17.5, 5.0)),
    ((1, 1.5), (17, 0.5)),
    ((17, 0.5), (1, 1.5)),

    # inside hole
    ((6.5, 6.5), (17, 0.5)),
    ((17, 0.5), (6.5, 6.5)),
]

TEST_DATA_GRID_ENV = [
    # ((start,goal),(path,distance))
    # shortest paths should be distinct (reverse will automatically be tested)

    # identical nodes
    (((15, 5), (15, 5)), ([(15, 5), (15, 5)], 0.0)),

    # directly reachable
    (((15, 5), (15, 6)), ([(15, 5), (15, 6)], 1.0)),
    (((15, 5), (16, 6)), ([(15, 5), (16, 6)], sqrt(2))),

    # points on the polygon edges (vertices) should be accepted!
    # on edge
    (((15, 0), (15, 6)), ([(15, 0), (15, 6)], 6.0)),
    (((17, 5), (16, 5)), ([(17, 5), (16, 5)], 1.0)),
    # on edge of hole
    (((7, 8), (7, 9)), ([(7, 8), (7, 9)], 1.0)),

    # on vertex
    (((4, 2), (4, 3)), ([(4, 2), (4, 3)], 1.0)),
    # on vertex of hole
    (((6, 8), (6, 9)), ([(6, 8), (6, 9)], 1.0)),

    # on two vertices
    # coinciding with edge (direct neighbour)
    (((4, 2), (4, 1)), ([(4, 2), (4, 1)], 1.0)),
    (((5, 5), (5, 7)), ([(5, 5), (5, 7)], 2.0)),
    # should have direct connection to all visible extremities! connected in graph
    (((6, 8), (5, 7)), ([(6, 8), (5, 7)], sqrt(2))),
    (((4, 1), (5, 7)), ([(4, 1), (5, 7)], sqrt(1 ** 2 + 6 ** 2))),
    # should have direct connection to all visible extremities! even if not connected in graph!
    (((4, 2), (5, 7)), ([(4, 2), (5, 7)], sqrt(1 ** 2 + 5 ** 2))),

    # mix of edges and vertices, directly visible
    (((2, 2), (5, 7)), ([(2, 2), (5, 7)], sqrt(3 ** 2 + 5 ** 2))),

    # also regular points should have direct connection to all visible extremities!
    (((10, 3), (17, 6)), ([(10, 3), (17, 6)], sqrt(7 ** 2 + 3 ** 2))),
    (((10, 3), (8, 8)), ([(10, 3), (8, 8)], sqrt(2 ** 2 + 5 ** 2))),
    # even if the query point lies in front of an extremity! (test if new query vertices are being created!)
    (((10, 3), (8, 5)), ([(10, 3), (8, 5)], sqrt(2 ** 2 + 2 ** 2))),

    # using a* graph search:
    # directly reachable through a single vertex (does not change distance!)
    (((5, 1), (3, 3)), ([(5, 1), (4, 2), (3, 3)], sqrt(2 ** 2 + 2 ** 2))),

    # If two Polygons have vertices with identical coordinates (this is allowed),
    #   paths through these vertices are theoretically possible!
    (((6.5, 5.5), (7.5, 6.5)), ([(6.5, 5.5), (7, 6), (7.5, 6.5)], sqrt(1 ** 2 + 1 ** 2))),

    # distance should stay the same even if multiple extremities lie on direct path
    # test if path is skipping passed extremities
    (((8, 4), (8, 8)), ([(8, 4), (8, 5), (8, 6), (8, 7), (8, 8)], 4)),
    (((8, 4), (8, 9)), ([(8, 4), (8, 5), (8, 6), (8, 7), (8, 8), (8, 9)], 5)),

    # regular examples
    (((0.5, 6), (18.5, 0.5)),
     ([(0.5, 6.0), (5, 5), (6, 5), (7, 5), (8, 5), (17, 6), (18, 6), (18.5, 0.5)], 23.18783787537749)),

    (((0.5, 6), (9, 5.5)),
     ([(0.5, 6.0), (5, 5), (6, 5), (7, 5), (8, 5), (9.0, 5.5)], 8.727806217396338)),

    (((0.5, 6), (18.5, 9)),
     ([(0.5, 6.0), (5, 5), (6, 5), (7, 5), (8, 5), (18, 7), (18.5, 9.0)], 19.869364068640845)),

    (((6.9, 4), (7, 9)),
     ([(6.9, 4.0), (7, 6), (8, 7), (8, 8), (7, 9)], 5.830925564196269)),

    (((6.5, 4), (7, 9)),
     ([(6.5, 4.0), (7, 6), (8, 7), (8, 8), (7, 9)], 5.889979937555021)),
]

POLY_ENV_PARAMS = (
    # boundary_coordinates
    [(0.0, 0.0), (10.0, 0.0), (9.0, 5.0), (10.0, 10.0), (0.0, 10.0)],
    # list_of_holes
    [[(3.0, 7.0), (5.0, 9.0), (4.6, 7.0), (5.0, 4.0), ], ],
)

TEST_DATA_POLY_ENV = [
    # ((start,goal),(path,distance))
    # identical nodes
    (((1, 1), (1, 1)), ([(1, 1), (1, 1)], 0.0)),

    # directly reachable
    (((1, 1), (1, 2)), ([(1, 1), (1, 2)], 1.0)),
    (((1, 1), (2, 1)), ([(1, 1), (2, 1)], 1.0)),

    # points on the polygon edges (vertices) should be accepted!
    # on edge (boundary polygon)
    (((1, 0), (1, 1)), ([(1, 0), (1, 1)], 1.0)),
    (((9.5, 2.5), (8.5, 2.5)), ([(9.5, 2.5), (8.5, 2.5)], 1.0)),
    (((0, 2), (0, 1)), ([(0, 2), (0, 1)], 1.0)),  # both
    (((1, 0), (5, 0)), ([(1, 0), (5, 0)], 4.0)),  # both
    # on edge of hole
    (((4, 8), (3, 8)), ([(4, 8), (3, 8)], 1.0)),
    (((4, 8), (4.1, 8.1)), ([(4, 8), (4.1, 8.1)], sqrt(2 * (0.1 ** 2)))),  # both

    # on vertex
    (((9, 5), (8, 5)), ([(9, 5), (8, 5)], 1.0)),
    # on vertex of hole
    (((3, 7), (2, 7)), ([(3, 7), (2, 7)], 1.0)),

    # on two vertices
    # coinciding with edge (direct neighbour)
    (((3, 7), (5, 9)), ([(3, 7), (5, 9)], sqrt(8))),
    (((4.6, 7), (5, 9)), ([(4.6, 7), (5, 9)], sqrt((0.4 ** 2) + (2 ** 2)))),
    # should have direct connection to all visible extremities! connected in graph
    (((5, 4), (5, 9)), ([(5, 4), (5, 9)], 5)),
    # should have a direct connection to all visible extremities! even if not connected in graph!
    (((9, 5), (5, 9)), ([(9, 5), (5, 9)], sqrt(2 * (4 ** 2)))),

    # using a* graph search:
    # directly reachable through a single vertex (does not change distance!)
    (((9, 4), (9, 6)), ([(9, 4), (9, 5), (9, 6)], 2)),
]


def try_test_cases(environment, test_cases):
    def validate(start_coordinates, goal_coordinates, expected_output):
        output = environment.find_shortest_path(start_coordinates, goal_coordinates)
        path, length = output
        expected_path, expected_length = expected_output
        correct_result = path == expected_path and length == pytest.approx(expected_length)
        if correct_result:
            status_str = 'OK'
        else:
            status_str = 'XX'
        print(f'{status_str} input: {(start_coordinates, goal_coordinates)} ')
        if PLOT_TEST_RESULTS:
            assert correct_result, \
                f'unexpected result (path, length): got {output} instead of {expected_output} '

    print('testing if path and distance are correct:')
    for ((start_coordinates, goal_coordinates), expected_output) in test_cases:
        validate(start_coordinates, goal_coordinates, expected_output)
        # automatically test reversed!
        path, length = expected_output
        expected_output_reversed = list(reversed(path)), length
        validate(goal_coordinates, start_coordinates, expected_output_reversed)


class MainTest(unittest.TestCase):

    def test_fct(self):
        grid_env = ENVIRONMENT_CLASS(**CONSTRUCTION_KWARGS)

        grid_env.store_grid_world(*GRID_ENV_PARAMS, simplify=False, validate=False)
        assert len(list(grid_env.all_extremities)) == 17, 'extremities do not get detected correctly!'
        grid_env.prepare()
        # raise ValueError
        assert len(grid_env.graph.all_nodes) == 16, 'identical nodes should get joined in the graph!'

        # test if points outside the map are being rejected
        for start_coordinates, goal_coordinates in INVALID_DESTINATION_DATA:
            with pytest.raises(ValueError):
                grid_env.find_shortest_path(start_coordinates, goal_coordinates)

        print('testing grid environment')
        try_test_cases(grid_env, TEST_DATA_GRID_ENV)

        # when the deep copy mechanism works correctly
        # even after many queries the internal graph should have the same structure as before
        # otherwise the temporarily added vertices during a query stay stored
        assert len(grid_env.graph.all_nodes) == 16, 'the graph should stay unchanged by shortest path queries!'

        nr_nodes_env1_old = len(grid_env.graph.all_nodes)

        poly_env = ENVIRONMENT_CLASS(**CONSTRUCTION_KWARGS)
        poly_env.store(*POLY_ENV_PARAMS, validate=True)
        NR_EXTR_POLY_ENV = 4
        assert len(list(poly_env.all_extremities)) == NR_EXTR_POLY_ENV, \
            f'the environment should detect all {NR_EXTR_POLY_ENV} extremities!'
        poly_env.prepare()
        nr_nodes_env2 = len(poly_env.graph.all_nodes)
        print(poly_env.graph.all_nodes)
        print(list(poly_env.all_extremities))
        assert nr_nodes_env2 == NR_EXTR_POLY_ENV, \
            f'the visibility graph should store all {NR_EXTR_POLY_ENV} extremities!'

        nr_nodes_env1_new = len(grid_env.graph.all_nodes)
        assert nr_nodes_env1_new == nr_nodes_env1_old, \
            'node amount of an grid_env should not change by creating another grid_env!'
        assert grid_env.graph is not poly_env.graph, \
            'different environments share the same graph object'
        assert grid_env.graph.all_nodes is not poly_env.graph.all_nodes, \
            'different environments share the same set of nodes'

        print('\ntesting polygon environment')
        try_test_cases(poly_env, TEST_DATA_POLY_ENV)

        # TODO test: When the paths should be blocked, use a single polygon with multiple identical
        #   vertices instead (also allowed?! change data requirements in doc!).

        # TODO test graph construction
        # when two nodes have the same angle representation there should only be an edge to the closer node!
        # test if property 1 is being properly exploited
        # (extremities lying in front of each other need not be connected)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(MainTest)
    unittest.TextTestRunner(verbosity=2).run(suite)
    unittest.main()



import heapq
from typing import List, Optional

import numpy as np

# TODO find a way to avoid global variable, wrap all in a different kind of 'coordinate system environment'?
# placeholder for temporarily storing the current origin
origin = None


class AngleRepresentation(object):
    """
    a class automatically computing a representation for the angle from the origin to a given vector
    value in [0.0 : 4.0[
    every quadrant contains angle measures from 0.0 to 1.0
    there are 4 quadrants (counter clockwise numbering)
    0 / 360 degree -> 0.0
    90 degree -> 1.0
    180 degree -> 2.0
    270 degree -> 3.0
    ...
    Useful for comparing angles without actually computing expensive trigonometrical functions
    This representation does not grow directly proportional to its represented angle,
    but it its bijective and monotonous:
    rep(p1) > rep(p2) <=> angle(p1) > angle(p2)
    rep(p1) = rep(p2) <=> angle(p1) = angle(p2)
    angle(p): counter clockwise angle between the two line segments (0,0)'--(1,0)' and (0,0)'--p
    with (0,0)' being the vector representing the origin

    """
    # prevent dynamic attribute assignment (-> safe memory)
    # __slots__ = ['quadrant', 'angle_measure', 'value']
    __slots__ = ['value']

    def __init__(self, np_vector):
        # 2D vector: (dx, dy) = np_vector
        norm = np.linalg.norm(np_vector)
        if norm == 0.0:
            # make sure norm is not 0!
            raise ValueError('received null vector:', np_vector, norm)

        dx_positive = np_vector[0] >= 0
        dy_positive = np_vector[1] >= 0

        if dx_positive and dy_positive:
            quadrant = 0.0
            angle_measure = np_vector[1] / norm

        elif not dx_positive and dy_positive:
            quadrant = 1.0
            angle_measure = -np_vector[0] / norm

        elif not dx_positive and not dy_positive:
            quadrant = 2.0
            angle_measure = -np_vector[1] / norm

        else:
            quadrant = 3.0
            angle_measure = np_vector[0] / norm

        self.value = quadrant + angle_measure

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return self.__str__()


class Vertex(object):
    # defining static attributes on class to safe memory
    __slots__ = ['coordinates', 'is_extremity', 'is_outdated', 'coordinates_translated', 'angle_representation',
                 'distance_to_origin']

    def __init__(self, coordinates):
        self.coordinates = np.array(coordinates)
        self.is_extremity: bool = False

        # a container for temporally storing shifted coordinates
        self.is_outdated: bool = True
        self.coordinates_translated = None
        self.angle_representation: Optional[AngleRepresentation] = None
        self.distance_to_origin: float = 0.0

    def __gt__(self, other):
        # ordering needed for priority queue. multiple vertices possibly have the same priority.
        return id(self) > id(other)

    def __str__(self):
        return str(self.coordinates)

    def __repr__(self):
        return self.__str__()

    def evaluate(self):
        global origin
        # store the coordinate value of the point relative to the new origin vector
        self.coordinates_translated = self.coordinates - origin.coordinates
        self.distance_to_origin = np.linalg.norm(self.coordinates_translated)
        if self.distance_to_origin == 0.0:
            # the coordinates of the origin and this vertex are equal
            # an angle is not defined in this case!
            self.angle_representation = None
        else:
            self.angle_representation = AngleRepresentation(self.coordinates_translated)

        self.is_outdated = False

    # lazy evaluation
    def get_coordinates_translated(self):
        if self.is_outdated:
            self.evaluate()
        return self.coordinates_translated

    def get_angle_representation(self):
        if self.is_outdated:
            self.evaluate()
        try:
            return self.angle_representation.value
        except AttributeError:
            return None

    def get_distance_to_origin(self):
        if self.is_outdated:
            self.evaluate()
        return self.distance_to_origin

    def mark_outdated(self):
        self.is_outdated = True


class PolygonVertex(Vertex):
    # __slots__ declared in parents are available in child classes. However, child subclasses will get a __dict__
    # and __weakref__ unless they also define __slots__ (which should only contain names of any additional slots).
    __slots__ = ['edge1', 'edge2', 'neighbour1', 'neighbour2']

    def __init__(self, *args, **kwargs):
        super(PolygonVertex, self).__init__(*args, **kwargs)
        self.edge1: Optional[Edge] = None
        self.edge2: Optional[Edge] = None
        self.neighbour1: Optional[PolygonVertex] = None
        self.neighbour2: Optional[PolygonVertex] = None

    def get_neighbours(self):
        return self.neighbour1, self.neighbour2

    def declare_extremity(self):
        self.is_extremity = True

    def set_edge1(self, e1):
        self.edge1 = e1
        # ordering is important! the numbering convention has to stay intact!
        self.neighbour1 = e1.vertex1

    def set_edge2(self, e2):
        self.edge2 = e2
        # ordering is important! the numbering convention has to stay intact!
        self.neighbour2 = e2.vertex2


class Edge(object):
    __slots__ = ['vertex1', 'vertex2']

    def __init__(self, vertex1, vertex2):
        self.vertex1: PolygonVertex = vertex1
        self.vertex2: PolygonVertex = vertex2

    def __str__(self):
        return self.vertex1.__str__() + '-->' + self.vertex2.__str__()

    def __repr__(self):
        return self.__str__()


class Polygon(object):
    __slots__ = ['vertices', 'edges', 'extremities', 'coordinates',
                 # 'is_hole', 'length',
                 ]

    def __init__(self, coordinate_list, is_hole):
        # store just the coordinates separately from the vertices in the format suiting the inside_polygon() function
        self.coordinates = np.array(coordinate_list)

        # self.is_hole: bool = is_hole
        # self.length: int = len(coordinate_list)

        if len(coordinate_list) < 3:
            raise ValueError('This is not a valid polygon:', coordinate_list, '# edges:', len(coordinate_list))

        self.vertices: List[PolygonVertex] = [PolygonVertex(coordinate) for coordinate in coordinate_list]

        self.edges: List[Edge] = []
        vertex1 = self.vertices[-1]
        for vertex2 in self.vertices:
            edge = Edge(vertex1, vertex2)
            # ordering is important! the numbering convention has to stay intact!
            vertex1.set_edge2(edge)
            vertex2.set_edge1(edge)
            self.edges.append(edge)
            vertex1 = vertex2

        def find_extremities():
            """
            identify all protruding points = vertices with an inside angle of > 180 degree ('extremities')
            expected edge numbering:
                outer boundary polygon: counter clockwise
                holes: clockwise
            :return:
            """
            self.extremities = []
            # extremity_indices = []
            # extremity_index = -1
            v1 = self.vertices[-2]
            p1 = v1.coordinates
            v2 = self.vertices[-1]
            p2 = v2.coordinates

            for v3 in self.vertices:

                p3 = v3.coordinates
                # since consequent vertices are not permitted to be equal,
                #   the angle representation of the difference is well defined
                if (AngleRepresentation(p3 - p2).value - AngleRepresentation(p1 - p2).value) % 4 < 2.0:
                    # basic idea:
                    #   - translate the coordinate system to have p2 as origin
                    #   - compute the angle representations of both vectors representing the edges
                    #   - "rotate" the coordinate system (equal to deducting) so that the p1p2 representation is 0
                    #   - check in which quadrant the p2p3 representation lies
                    # %4 because the quadrant has to be in [0,1,2,3] (representation in [0:4[)
                    # if the representation lies within quadrant 0 or 1 (<2.0), the inside angle
                    #   (for boundary polygon inside, for holes outside) between p1p2p3 is > 180 degree
                    # then p2 = extremity
                    v2.declare_extremity()
                    self.extremities.append(v2)

                # move to the next point
                # vertex1=vertex2
                v2 = v3
                p1 = p2
                p2 = p3

        self.extremities: List[PolygonVertex] = None
        find_extremities()

    def translate(self, new_origin: Vertex):
        global origin
        origin = new_origin
        for vertex in self.vertices:
            vertex.mark_outdated()


class PriorityQueue:
    def __init__(self):
        self.elements = []

    def empty(self):
        return len(self.elements) == 0

    def put(self, item, priority):
        heapq.heappush(self.elements, (priority, item))

    def get(self):
        return heapq.heappop(self.elements)[1]  # return only the item without the priority


# TODO often empty sets in self.neighbours
class DirectedHeuristicGraph(object):
    __slots__ = ['all_nodes', 'distances', 'goal_node', 'heuristic', 'neighbours']

    def __init__(self):
        self.distances: dict = {}
        self.neighbours: dict = {}
        self.all_nodes: set = set()
        # TODO use same set as extremities of env, but different for copy!

        # the heuristic must NEVER OVERESTIMATE the actual cost (here: actual shortest distance)
        # <=> must always be lowest for node with the POSSIBLY lowest cost
        # <=> heuristic is LOWER BOUND for the cost
        # the heuristic here: distance to the goal node (is always the shortest possible distance!)
        self.heuristic: dict = {}
        self.goal_node: Optional[Vertex] = None

    def __deepcopy__(self, memodict=None):
        # returns an independent copy (nodes can be added without changing the original graph),
        # but without actually copying vertex instances!
        independent_copy = DirectedHeuristicGraph()
        independent_copy.distances = self.distances.copy()
        independent_copy.neighbours = {k: v.copy() for k, v in self.neighbours.items()}
        independent_copy.all_nodes = self.all_nodes.copy()
        return independent_copy

    def get_all_nodes(self):
        return self.all_nodes

    def get_neighbours(self):
        return self.neighbours.items()

    def get_neighbours_of(self, node):
        return self.neighbours.get(node, set())

    def get_distance(self, node1, node2):
        # directed edges: just one direction is being stored
        return self.distances[(node1, node2)]

    def get_heuristic(self, node):
        global origin  # use origin contains the current goal node
        # lazy evaluation:
        h = self.heuristic.get(node, None)
        if h is None:
            # has been reset, compute again
            h = np.linalg.norm(node.coordinates - origin.coordinates)
            self.heuristic[node] = h
        return h

    def set_goal_node(self, goal_node):
        assert goal_node in self.all_nodes  # has no outgoing edges -> no neighbours
        # IMPORTANT: while using heuristic graph (a star), do not change origin!
        global origin
        origin = goal_node  # use origin for temporally storing the goal node
        self.goal_node = goal_node
        # reset heuristic for all
        self.heuristic.clear()

    def edges_from(self, node1):
        # optimisation:
        #   return the neighbours ordered after their cost estimate: distance+ heuristic (= current-next + next-goal)
        #   -> when the goal is reachable return it first (-> a star search terminates)

        neighbours = self.get_neighbours_of(node1)
        distances = [self.get_distance(node1, n) for n in neighbours]
        out_sorted = sorted([(n, distances[i], distances[i] + self.get_heuristic(n)) for i, n in enumerate(neighbours)],
                            key=lambda x: x[2])

        # yield node, distance, cost= distance + heuristic
        yield from out_sorted

    def add_directed_edge(self, node1, node2, distance):
        assert node1 != node2  # no self loops allowed!
        self.neighbours.setdefault(node1, set()).add(node2)
        self.distances[(node1, node2)] = distance
        self.all_nodes.add(node1)
        self.all_nodes.add(node2)

    def add_undirected_edge(self, node1, node2, distance):
        assert node1 != node2  # no self loops allowed!
        self.add_directed_edge(node1, node2, distance)
        self.add_directed_edge(node2, node1, distance)

    def add_multiple_undirected_edges(self, node1, node_distance_iter):
        for node2, distance in node_distance_iter:
            self.add_undirected_edge(node1, node2, distance)

    def add_multiple_directed_edges(self, node1, node_distance_iter):
        for node2, distance in node_distance_iter:
            self.add_directed_edge(node1, node2, distance)

    def remove_directed_edge(self, n1, n2):
        neighbours = self.neighbours.get(n1)
        if neighbours is not None:
            neighbours.discard(n2)
            self.distances.pop((n1, n2), None)
            # ATTENTION: even if there are no neighbours left and a node is hence dangling (not reachable),
            # the node must still be kept, since with the addition of start and goal nodes during a query
            # the node might become reachable!

    def remove_undirected_edge(self, node1, node2):
        # should work even if edge does not exist yet
        self.remove_directed_edge(node1, node2)
        self.remove_directed_edge(node2, node1)

    def remove_multiple_undirected_edges(self, node1, node2_iter):
        for node2 in node2_iter:
            self.remove_undirected_edge(node1, node2)

    def make_clean(self):
        # for shortest path computations all graph nodes should be unique
        self.join_identical()
        # leave dangling nodes! (they might become reachable by adding start and and goal node!)

    def join_identical(self):
        # join all nodes with the same coordinates,
        nodes_to_check = self.get_all_nodes().copy()
        while len(nodes_to_check) > 1:
            n1 = nodes_to_check.pop()
            coordinates1 = n1.coordinates
            same_nodes = {n for n in nodes_to_check if np.allclose(coordinates1, n.coordinates)}
            nodes_to_check.difference_update(same_nodes)
            for n2 in same_nodes:
                # print('removing duplicate node', n2)
                neighbours_n1 = self.neighbours[n1]
                neighbours_n2 = self.neighbours.pop(n2)
                for n3 in neighbours_n2:
                    d = self.distances.pop((n2, n3))
                    self.distances.pop((n3, n2))
                    self.neighbours[n3].remove(n2)
                    # do not allow self loops!
                    if n3 != n1 and n3 not in neighbours_n1:
                        # and add all the new edges to node 1
                        self.add_undirected_edge(n1, n3, d)

                self.all_nodes.remove(n2)

    def modified_a_star(self, start, goal):
        """ implementation of the popular A* algorithm with optimisations for this special use case

        IMPORTANT: geometrical property of this problem (and hence also the extracted graph):
        it is always shortest to directly reach a node instead of visiting other nodes first
        (there is never an advantage through reduced edge weight)

        this can be exploited in a lot of cases to make A* faster and terminate earlier than for general graphs:
        -> when the goal is directly reachable, there can be no other shorter path to it. terminate.
        -> there is no need to revisit nodes (path only gets longer)
        -> not all neighbours of the current node have to be checked like in vanilla A* before continuing
        to the next node.

        Optimisation: keep one 'neighbour generator' open for every visited node,
            that is yielding its neighbours starting with the one having the lowest cost estimate.
            One then only has to store those generators in a priority queue
            to always draw the generator with the neighbour having the lowest TOTAL cost estimate (lower bound) next.

        modified sample code from https://www.redblobgames.com/pathfinding/a-star/

        :param start: the vertex to start from
        :param goal: the vertex to end at
        :return: a tuple of the shortest path from start to goal and its total length
        """

        def enqueue_next_from(generator):
            try:
                # node, distance, cost estimate (= distance + heuristic = current-next + estimate(next-goal))
                next_n, dist, cost_estim = (next(generator))
                # the priority has to be the lower bound (=estimate) of the TOTAL cost!
                # = cost_so_far + cost_estim  (= start-current + estimate(current-goal))
                priority_queue.put((next_n, dist, generator, path, cost_so_far), cost_so_far + cost_estim)
            except StopIteration:
                # there is no neighbour left
                pass

        self.set_goal_node(goal)  # lazy update of the heuristic

        priority_queue = PriorityQueue()
        neighbour_gen = self.edges_from(start)
        cost_so_far = 0.0
        path = [start]
        enqueue_next_from(neighbour_gen)
        visited_nodes = set()

        while not priority_queue.empty():
            # always 'visit' the node with the current lowest total cost estimate
            current, distance, neighbour_gen, path, cost_so_far = priority_queue.get()
            # print('visiting:', current)
            # print('neighbours:', heuristic_graph.get_neighbours_of(current))

            # there could still be other neighbours left in this generator:
            enqueue_next_from(neighbour_gen)

            if current in visited_nodes:
                # this node has already been visited. there is no need to consider
                # path can only get longer by visiting other nodes first: new_cost is never < cost_so_far
                continue

            visited_nodes.add(current)
            # NOTE: in contrast to vanilla A*,
            # here the path and cost have to be stored separately for every open generator!
            cost_so_far += distance
            # IMPORTANT: use independent path lists
            path = path.copy()
            path.append(current)

            if current == goal:
                # since the goal node is the one with the lowest cost estimate
                # because of the geometric property mentioned above there can be no other shortest path to the goal
                # the algorithm can be terminated (regular a* would now still continue to all other neighbours first)
                # optimisation: _exiting_edges_from() returns the goal node first if it is among the neighbours
                # heuristic(goal) == 0
                # print('reached goal node. terminating')
                return path, cost_so_far

            # add neighbours of the current node
            neighbour_gen = self.edges_from(current)
            enqueue_next_from(neighbour_gen)

        # goal is not reachable
        return [], None



import unittest

import numpy as np

from extremitypathfinder.helper_classes import AngleRepresentation
from extremitypathfinder.helper_fcts import has_clockwise_numbering, inside_polygon
from helpers import proto_test_case


class HelperFctsTest(unittest.TestCase):

    def value_test_fct(input):
        np_2D_coord_vector = np.array(input)
        return AngleRepresentation(np_2D_coord_vector).value

    data = [
        ([1.0, 0.0], 0.0),
        ([0.0, 1.0], 1.0),
        ([-1.0, 0.0], 2.0),
        ([0.0, -1.0], 3.0),

        ([2.0, 0.0], 0.0),
        ([0.0, 2.0], 1.0),
        ([-2.0, 0.0], 2.0),
        ([0.0, -2.0], 3.0),
    ]

    proto_test_case(data, value_test_fct)

    def test_inside_polygon(self):
        # TODO more detailed test. edges with slopes... also in timezonefinder

        for border_value in [True, False]:
            def test_fct(input):
                polygon_test_case = np.array([(-1.0, -1.0), (1.0, -1.0), (1.0, 1.0), (-1.0, 1.0)])
                x, y = input

                return inside_polygon(x, y, polygon_test_case, border_value)

            p_test_cases = [
                # (x,y),
                # inside
                (0.0, 0.0),
                # # outside
                (-2.0, 2.0),
                (0, 2.0),
                (2.0, 2.0),
                (-2.0, 0),
                (2.0, 0),
                (-2.0, -2.0),
                (0, -2.0),
                (2.0, -2.0),

                # on the line test cases
                (-1.0, -1.0), (1.0, -1.0), (1.0, 1.0), (-1.0, 1.0),
                (0.0, 1),
                (0, -1),
                (1, 0),
                (-1, 0),
            ]
            expected_results = [
                True, False, False, False, False, False, False, False, False,
                # on the line test cases
                border_value, border_value, border_value, border_value, border_value, border_value, border_value,
                border_value,
            ]

            proto_test_case(list(zip(p_test_cases, expected_results)), test_fct)

    def test_clockwise_numering(self):
        def clockwise_test_fct(input):
            return has_clockwise_numbering(np.array(input))

        data = [
            # clockwise numbering!
            ([(3.0, 7.0), (5.0, 9.0), (5.0, 7.0)], True),
            ([(3.0, 7.0), (5.0, 9.0), (4.5, 7.0), (5.0, 4.0)], True),
            ([(0.0, 0.0), (0.0, 1.0), (1.0, 0.0)], True),

            # # counter clockwise edge numbering!
            ([(0.0, 0.0), (1.0, 0.0), (0.0, 1.0)], False),
            ([(0.0, 0.0), (10.0, 0.0), (10.0, 10.0), (0.0, 10.0)], False),
            ([(0.0, 0.0), (10.0, 0.0), (10.0, 5.0), (10.0, 10.0), (0.0, 10.0)], False),
            ([(0.0, 0.0), (10.0, 0.0), (9.0, 5.0), (10.0, 10.0), (0.0, 10.0)], False),

        ]
        proto_test_case(data, clockwise_test_fct)


# TODO test if relation is really bidirectional (y in find_visible(x,y) <=> x in find_visible(y,x))


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(HelperFctsTest)
    unittest.TextTestRunner(verbosity=2).run(suite)
    # unittest.main()



import pickle
from copy import deepcopy
from typing import Iterable, List, Tuple, Union

import numpy as np

from extremitypathfinder.helper_classes import DirectedHeuristicGraph, Edge, Polygon, PolygonVertex, Vertex
from extremitypathfinder.helper_fcts import (
    check_data_requirements, convert_gridworld, find_visible, find_within_range, inside_polygon,
)

# TODO possible to allow polygon consisting of 2 vertices only(=barrier)? lots of functions need at least 3 vertices atm

COORDINATE_TYPE = Tuple[float, float]
PATH_TYPE = List[COORDINATE_TYPE]
LENGTH_TYPE = float
INPUT_NUMERICAL_TYPE = Union[float, int]
INPUT_COORD_TYPE = Tuple[INPUT_NUMERICAL_TYPE, INPUT_NUMERICAL_TYPE]
OBSTACLE_ITER_TYPE = Iterable[INPUT_COORD_TYPE]
INPUT_COORD_LIST_TYPE = Union[np.ndarray, List]

DEFAULT_PICKLE_NAME = 'environment.pickle'


# is not a helper function to make it an importable part of the package
def load_pickle(path=DEFAULT_PICKLE_NAME):
    print('loading map from:', path)
    with open(path, 'rb') as f:
        return pickle.load(f)


# TODO document parameters
class PolygonEnvironment:
    """ class allowing to use polygons to represent "2D environments" and use them for path finding.

    keeps a "loaded" and prepared environment for consecutive path queries.
    internally uses a visibility graph optimised for shortest path finding.
    general approach and some optimisations theoretically described in:
    [1] Vinther, Anders Strand-Holm, Magnus Strand-Holm Vinther, and Peyman Afshani.
    "`Pathfinding in Two-dimensional Worlds
    <https://www.cs.au.dk/~gerth/advising/thesis/anders-strand-holm-vinther_magnus-strand-holm-vinther.pdf>`__"
    """

    boundary_polygon: Polygon = None
    holes: List[Polygon] = None
    prepared: bool = False
    graph: DirectedHeuristicGraph = None
    temp_graph: DirectedHeuristicGraph = None  # for storing and plotting the graph during a query

    @property
    def polygons(self) -> Iterable[Polygon]:
        yield self.boundary_polygon
        yield from self.holes

    @property
    def all_vertices(self) -> List[PolygonVertex]:
        for p in self.polygons:
            yield from p.vertices

    @property
    def all_extremities(self) -> Iterable[PolygonVertex]:
        for p in self.polygons:
            yield from p.extremities

    @property
    def all_edges(self) -> Iterable[Edge]:
        for p in self.polygons:
            yield from p.edges

    def store(self, boundary_coordinates: INPUT_COORD_LIST_TYPE, list_of_hole_coordinates: INPUT_COORD_LIST_TYPE,
              validate: bool = False):
        """ saves the passed input polygons in the environment

        .. note:: the passed polygons must meet these requirements:

            * given as numpy or python array of coordinate tuples: ``[(x1,y1), (x2,y2,)...]``
            * no repetition of the first point at the end
            * at least 3 vertices (no single points or lines allowed)
            * no consequent vertices with identical coordinates in the polygons (same coordinates allowed)
            * no self intersections
            * no true intersections with other polygons, identical vertices allowed
            * edge numbering has to follow these conventions: boundary polygon counter clockwise, holes clockwise

        :param boundary_coordinates: array of coordinates with counter clockwise edge numbering
        :param list_of_hole_coordinates: array of coordinates with clockwise edge numbering
        :param validate: whether the requirements of the data should be tested
        """
        self.prepared = False
        # 'loading the map
        boundary_coordinates = np.array(boundary_coordinates)
        list_of_hole_coordinates = [np.array(hole_coords) for hole_coords in list_of_hole_coordinates]
        if validate:
            check_data_requirements(boundary_coordinates, list_of_hole_coordinates)

        self.boundary_polygon = Polygon(boundary_coordinates, is_hole=False)
        # IMPORTANT: make a copy of the list instead of linking to the same list (python!)
        self.holes = [Polygon(coordinates, is_hole=True) for coordinates in list_of_hole_coordinates]

    def store_grid_world(self, size_x: int, size_y: int, obstacle_iter: OBSTACLE_ITER_TYPE, simplify: bool = True,
                         validate: bool = False):
        """ convert a grid-like into a polygon environment and save it

        prerequisites: grid world must not have single non-obstacle cells which are surrounded by obstacles
        ("white cells in black surrounding" = useless for path planning)

        :param size_x: the horizontal grid world size
        :param size_y: the vertical grid world size
        :param obstacle_iter: an iterable of coordinate pairs (x,y) representing blocked grid cells (obstacles)
        :param validate: whether the input should be validated
        :param simplify: whether the polygons should be simplified or not. reduces edge amount, allow diagonal edges
        """
        boundary_coordinates, list_of_hole_coordinates = convert_gridworld(size_x, size_y, obstacle_iter, simplify)
        self.store(boundary_coordinates, list_of_hole_coordinates, validate)

    def export_pickle(self, path: str = DEFAULT_PICKLE_NAME):
        print('storing map class in:', path)
        with open(path, 'wb') as f:
            pickle.dump(self, f)
        print('done.\n')

    def translate(self, new_origin: Vertex):
        """ shifts the coordinate system to a new origin

        computing the angle representations, shifted coordinates and distances for all vertices
        respective to the query point (lazy!)

        :param new_origin: the origin of the coordinate system to be shifted to
        """
        for p in self.polygons:
            p.translate(new_origin)

    def prepare(self):
        """ computes a visibility graph optimized (=reduced) for path planning and stores it

        computes all directly reachable extremities based on visibility and their distance to each other

        .. note::
            multiple polygon vertices might have identical coordinates.
            they must be treated as distinct vertices here, since their attached edges determine visibility
            in the created graph however these nodes must be merged at the end to avoid ambiguities!

        .. note::
            pre computing the shortest paths between all directly reachable extremities
            and storing them in the graph would not be an advantage, because then the graph is fully connected
            a star would visit every node in the graph at least once (-> disadvantage!).
        """

        if self.prepared:
            raise ValueError('this environment is already prepared. load new polygons first.')

        # preprocessing the map
        # construct graph of visible (=directly reachable) extremities
        # and optimize graph further at construction time
        self.graph = DirectedHeuristicGraph()
        extremities_to_check = set(self.all_extremities)
        # have to run for all (also last one!), because existing edges might get deleted every loop
        while len(extremities_to_check) > 0:
            # extremities are always visible to each other (bi-directional relation -> undirected graph)
            #  -> do not check extremities which have been checked already
            #  (would only give the same result when algorithms are correct)
            # the extremity itself must not be checked when looking for visible neighbours
            query_extremity: PolygonVertex = extremities_to_check.pop()
            self.translate(new_origin=query_extremity)

            visible_vertices = set()
            candidate_extremities = extremities_to_check.copy()
            # remove the extremities with the same coordinates as the query extremity
            candidate_extremities.difference_update(
                {c for c in candidate_extremities if c.get_angle_representation() is None})

            # these vertices all belong to a polygon
            # direct neighbours of the query vertex are visible
            # neighbouring vertices are reachable with the distance equal to the edge length
            n1, n2 = query_extremity.get_neighbours()
            try:
                candidate_extremities.remove(n1)
                visible_vertices.add((n1, n1.get_distance_to_origin()))
            except KeyError:
                pass
            try:
                candidate_extremities.remove(n2)
                visible_vertices.add((n2, n2.get_distance_to_origin()))
            except KeyError:
                pass

            # even though candidate_extremities might be empty now
            # must not skip to next loop here, because existing graph edges might get deleted here!

            # eliminate all vertices 'behind' the query point from the candidate set
            # since the query vertex is an extremity the 'outer' angle is < 180 degree
            # then the difference between the angle representation of the two edges has to be < 2.0
            # all vertices between the angle of the two neighbouring edges ('outer side')
            #   are not visible (no candidates!)
            # vertices with the same angle representation might be visible! do not delete them!
            repr1 = n1.get_angle_representation()
            repr2 = n2.get_angle_representation()
            repr_diff = abs(repr1 - repr2)
            candidate_extremities.difference_update(
                find_within_range(repr1, repr2, repr_diff, candidate_extremities, angle_range_less_180=True,
                                  equal_repr_allowed=False))

            # as shown in [1, Ch. II 4.4.2 "Property One"] Starting from any point lying "in front of" an extremity e,
            # such that both adjacent edges are visible, one will never visit e, because everything is
            # reachable on a shorter path without e (except e itself). An extremity e1 lying in the area "in front of"
            #   extremity e hence is never the next vertex in a shortest path coming from e.
            #   And also in reverse: when coming from e1 everything else than e itself can be reached faster
            #   without visiting e2. -> e1 and e do not have to be connected in the graph.
            # IMPORTANT: this condition only holds for building the basic visibility graph!
            #   when a query point happens to be an extremity, edges to the (visible) extremities in front
            #   MUST be added to the graph!
            # find extremities which fulfill this condition for the given query extremity
            repr1 = (repr1 + 2.0) % 4.0  # rotate 180 deg
            repr2 = (repr2 + 2.0) % 4.0
            # IMPORTANT: the true angle diff does not change, but the repr diff does! compute again
            repr_diff = abs(repr1 - repr2)

            # IMPORTANT: check all extremities here, not just current candidates
            # do not check extremities with equal coordinates (also query extremity itself!)
            #   and with the same angle representation (those edges must not get deleted from graph!)
            temp_candidates = set(filter(lambda e: e.get_angle_representation() is not None, self.all_extremities))
            lie_in_front = find_within_range(repr1, repr2, repr_diff, temp_candidates, angle_range_less_180=True,
                                             equal_repr_allowed=False)

            # already existing edges in the graph to the extremities in front have to be removed
            self.graph.remove_multiple_undirected_edges(query_extremity, lie_in_front)
            # do not consider when looking for visible extremities (NOTE: they might actually be visible!)
            candidate_extremities.difference_update(lie_in_front)

            # all edges except the neighbouring edges (handled above!) have to be checked
            edges_to_check = set(self.all_edges)
            edges_to_check.remove(query_extremity.edge1)
            edges_to_check.remove(query_extremity.edge2)

            visible_vertices.update(find_visible(candidate_extremities, edges_to_check))
            self.graph.add_multiple_undirected_edges(query_extremity, visible_vertices)

        # join all nodes with the same coordinates
        self.graph.make_clean()
        self.prepared = True

    # make sure start and goal are within the boundary polygon and outside of all holes
    def within_map(self, coords: INPUT_COORD_TYPE):
        """ checks if the given coordinates lie within the boundary polygon and outside of all holes

        :param coords: numerical tuple representing coordinates
        :return: whether the given coordinate is a valid query point
        """
        #
        x, y = coords
        if not inside_polygon(x, y, self.boundary_polygon.coordinates, border_value=True):
            return False
        for hole in self.holes:
            if inside_polygon(x, y, hole.coordinates, border_value=False):
                return False
        return True

    def find_shortest_path(self, start_coordinates: INPUT_COORD_TYPE, goal_coordinates: INPUT_COORD_TYPE,
                           free_space_after: bool = True, verify: bool = True) -> Tuple[PATH_TYPE, LENGTH_TYPE]:
        """ computes the shortest path and its length between start and goal node

        :param start_coordinates: a (x,y) coordinate tuple representing the start node
        :param goal_coordinates:  a (x,y) coordinate tuple representing the goal node
        :param free_space_after: whether the created temporary search graph self.temp_graph
            should be deleted after the query
        :param verify: whether it should be checked if start and goal points really lie inside the environment.
         if points close to or on polygon edges should be accepted as valid input, set this to ``False``.
        :return: a tuple of shortest path and its length
        """
        # path planning query:
        # make sure the map has been loaded and prepared
        if self.boundary_polygon is None:
            raise ValueError('No Polygons have been loaded into the map yet.')
        if not self.prepared:
            self.prepare()

        if verify and not (self.within_map(start_coordinates) and self.within_map(goal_coordinates)):
            raise ValueError('start or goal do not lie within the map')

        if start_coordinates == goal_coordinates:
            # start and goal are identical and can be reached instantly
            return [start_coordinates, goal_coordinates], 0.0

        # could check if start and goal nodes have identical coordinates with one of the vertices
        # optimisations for visibility test can be made in this case:
        # for extremities the visibility has already been (except for in front) computed
        # BUT: too many cases possible: e.g. multiple vertices identical to query point...
        # -> always create new query vertices
        # include start and goal vertices in the graph
        start_vertex = Vertex(start_coordinates)
        goal_vertex = Vertex(goal_coordinates)

        # check the goal node first (earlier termination possible)
        self.translate(new_origin=goal_vertex)  # do before checking angle representations!
        # IMPORTANT: manually translate the start vertex, because it is not part of any polygon
        #   and hence does not get translated automatically
        start_vertex.mark_outdated()

        # the visibility of only the graphs nodes has to be checked (not all extremities!)
        # points with the same angle representation should not be considered visible
        # (they also cause errors in the algorithms, because their angle repr is not defined!)
        candidates = set(filter(lambda n: n.get_angle_representation() is not None, self.graph.get_all_nodes()))
        # IMPORTANT: check if the start node is visible from the goal node!
        candidates.add(start_vertex)

        visibles_n_distances_goal = find_visible(candidates, edges_to_check=set(self.all_edges))
        if len(visibles_n_distances_goal) == 0:
            # The goal node does not have any neighbours. Hence there is not possible path to the goal.
            return [], None

        # create temporary graph TODO make more performant, avoid real copy
        # DirectedHeuristicGraph implements __deepcopy__() to not change the original precomputed self.graph
        # but to still not create real copies of vertex instances!
        self.temp_graph = deepcopy(self.graph)

        # IMPORTANT geometrical property of this problem: it is always shortest to directly reach a node
        #   instead of visiting other nodes first (there is never an advantage through reduced edge weight)
        # -> when goal is directly reachable, there can be no other shorter path to it. Terminate

        for v, d in visibles_n_distances_goal:
            if v == start_vertex:
                return [start_coordinates, goal_coordinates], d

            # add unidirectional edges to the temporary graph
            # add edges in the direction: extremity (v) -> goal
            # TODO: improvement: add edges last, after filtering them. instead of deleting edges
            self.temp_graph.add_directed_edge(v, goal_vertex, d)

        self.translate(new_origin=start_vertex)  # do before checking angle representations!
        # the visibility of only the graphs nodes have to be checked
        # the goal node does not have to be considered, because of the earlier check
        candidates = set(filter(lambda n: n.get_angle_representation() is not None, self.graph.get_all_nodes()))
        visibles_n_distances_start = find_visible(candidates, edges_to_check=set(self.all_edges))
        if len(visibles_n_distances_start) == 0:
            # The start node does not have any neighbours. Hence there is not possible path to the goal.
            return [], None

        # add edges in the direction: start -> extremity
        # TODO: improvement: add edges last, after filtering them. instead of deleting edges
        self.temp_graph.add_multiple_directed_edges(start_vertex, visibles_n_distances_start)

        # also here unnecessary edges in the graph can be deleted when start or goal lie in front of visible extremities
        # IMPORTANT: when a query point happens to coincide with an extremity, edges to the (visible) extremities
        #  in front MUST be added to the graph! Handled by always introducing new (non extremity, non polygon) vertices.

        # for every extremity that is visible from either goal or start
        # NOTE: edges are undirected! self.temp_graph.get_neighbours_of(start_vertex) == set()
        # neighbours_start = self.temp_graph.get_neighbours_of(start_vertex)
        neighbours_start = {n for n, d in visibles_n_distances_start}
        # the goal vertex might be marked visible, it is not an extremity -> skip
        neighbours_start.discard(goal_vertex)
        neighbours_goal = self.temp_graph.get_neighbours_of(goal_vertex)
        for vertex in neighbours_start | neighbours_goal:
            # assert type(vertex) == PolygonVertex and vertex.is_extremity

            # check only if point is visible
            temp_candidates = set()
            if vertex in neighbours_start:
                temp_candidates.add(start_vertex)

            if vertex in neighbours_goal:
                temp_candidates.add(goal_vertex)

            if len(temp_candidates) > 0:
                self.translate(new_origin=vertex)
                # IMPORTANT: manually translate the goal and start vertices
                start_vertex.mark_outdated()
                goal_vertex.mark_outdated()

                n1, n2 = vertex.get_neighbours()
                repr1 = (n1.get_angle_representation() + 2.0) % 4.0  # rotated 180 deg
                repr2 = (n2.get_angle_representation() + 2.0) % 4.0
                repr_diff = abs(repr1 - repr2)

                # IMPORTANT: special case:
                # here the nodes must stay connected if they have the same angle representation!
                lie_in_front = find_within_range(repr1, repr2, repr_diff, temp_candidates, angle_range_less_180=True,
                                                 equal_repr_allowed=False)
                self.temp_graph.remove_multiple_undirected_edges(vertex, lie_in_front)

        # NOTE: exploiting property 2 from [1] here would be more expensive than beneficial
        vertex_path, distance = self.temp_graph.modified_a_star(start_vertex, goal_vertex)

        if free_space_after:
            del self.temp_graph  # free the memory

        # extract the coordinates from the path
        return [tuple(v.coordinates) for v in vertex_path], distance


if __name__ == "__main__":
    # TODO command line support. read polygons and holes from .json files?
    pass



# -*- coding:utf-8 -*-

from setuptools import setup

setup(
    name='extremitypathfinder',
    packages=['extremitypathfinder'],
    description='python package for geometric shortest path computation for given 2D multi-polygon maps',
    # version: in VERSION file https://packaging.python.org/guides/single-sourcing-package-version/
    # With this approach you must make sure that the VERSION file is included in all your source
    # and binary distributions (e.g. add include VERSION to your MANIFEST.in).
    author='J. Michelfeit',
    author_email='python@michelfe.it',
    license='MIT licence',
    url='https://github.com/MrMinimal64/extremitypathfinder',  # use the URL to the github repo
    project_urls={
        "Source Code": "https://github.com/MrMinimal64/extremitypathfinder",
        "Documentation": "https://github.com/MrMinimal64/extremitypathfinder/blob/master/README.rst",
        "Changelog": "https://github.com/MrMinimal64/extremitypathfinder/blob/master/CHANGELOG.rst",
    },
    keywords='path-planning path-finding shortest-path visibility graph visibility-graph polygon'
             'robotics navigation offline ',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    long_description='python package for fast geometric shortest path computation in 2D multi-polygon '
                     'or grid environments based on visibility graphs.'
                     'Please refer to the `documentation <https://extremitypathfinder.readthedocs.io/en/latest/>`__.',
    python_requires='>=3.7',
    install_requires=[
        'numpy>=1.16',
    ],
)



from itertools import combinations
from typing import List

import numpy as np

from extremitypathfinder.helper_classes import AngleRepresentation, PolygonVertex


# TODO numba precompilation of some parts possible?! do line speed profiling first! speed impact
def inside_polygon(x, y, coords, border_value):
    # should return the border value for point equal to any polygon vertex
    # TODO overflow possible with large values when comparing slopes, change procedure
    for c in coords[:]:
        if np.all(c == [x, y]):
            return border_value

    # and if the point p lies on any polygon edge
    p = np.array([x, y])
    p1 = coords[-1, :]
    for p2 in coords[:]:
        if abs((AngleRepresentation(p1 - p).value - AngleRepresentation(p2 - p).value)) == 2.0:
            return border_value
        p1 = p2

    contained = False
    # the edge from the last to the first point is checked first
    i = -1
    y1 = coords[-1, 1]
    y_gt_y1 = y > y1
    for y2 in coords[:, 1]:
        y_gt_y2 = y > y2
        if y_gt_y1:
            if not y_gt_y2:
                x1 = coords[i, 0]
                x2 = coords[i + 1, 0]
                # only crossings "right" of the point should be counted
                x1GEx = x <= x1
                x2GEx = x <= x2
                # compare the slope of the line [p1-p2] and [p-p2]
                # depending on the position of p2 this determines whether the polygon edge is right or left of the point
                # to avoid expensive division the divisors (of the slope dy/dx) are brought to the other side
                # ( dy/dx > a  ==  dy > a * dx )
                if (x1GEx and x2GEx) or ((x1GEx or x2GEx) and (y2 - y) * (x2 - x1) <= (y2 - y1) * (x2 - x)):
                    contained = not contained

        else:
            if y_gt_y2:
                x1 = coords[i, 0]
                x2 = coords[i + 1, 0]
                # only crossings "right" of the point should be counted
                x1GEx = x <= x1
                x2GEx = x <= x2
                if (x1GEx and x2GEx) or ((x1GEx or x2GEx) and (y2 - y) * (x2 - x1) >= (y2 - y1) * (x2 - x)):
                    contained = not contained

        y1 = y2
        y_gt_y1 = y_gt_y2
        i += 1

    return contained


def no_identical_consequent_vertices(coords):
    p1 = coords[-1]
    for p2 in coords:
        # TODO adjust allowed difference epsilon
        assert not np.allclose(p1, p2)
        p1 = p2

    return True


def get_intersection_status(p1, p2, q1, q2):
    # return:
    #   0: no intersection
    #   1: intersection in ]p1;p2[
    # TODO 4 different possibilities
    #   2: intersection directly in p1 or p2
    #   3: intersection directly in q1 or q2
    # solve the set of equations
    # (p2-p1) lambda + (p1) = (q2-q1) mu + (q1)
    #  in matrix form A x = b:
    # [(p2-p1) (q1-q2)] (lambda, mu)' = (q1-p1)
    A = np.array([p2 - p1, q1 - q2]).T
    b = np.array(q1 - p1)
    try:
        x = np.linalg.solve(A, b)
        # not crossing the line segment is considered to be ok
        # so x == 0.0 or x == 1.0 is not considered an intersection
        # assert np.allclose((p2 - p1) * x[0] + p1, (q2 - q1) * x[1] + q1)
        # assert np.allclose(np.dot(A, x), b)
        if x[0] <= 0.0 or x[1] <= 0.0 or x[0] >= 1.0 or x[1] >= 1.0:
            return 0
        # if np.all(0.0 <= x) and np.all(x <= 1.0):
        #     return 2
    except np.linalg.LinAlgError:
        # line segments are parallel (matrix is singular, set of equations is not solvable)
        return 0

    return 1


# special case of has_intersection()
def lies_behind(p1, p2, v):
    # solve the set of equations
    # (p2-p1) lambda + (p1) = (v) mu
    #  in matrix form A x = b:
    # [(p1-p2) (v)] (lambda, mu)' = (p1)
    # because the vertex lies within the angle range between the two edge vertices
    #    (together with the other conditions on the polygons)
    #   this set of linear equations is always solvable (the matrix is regular)
    A = np.array([p1 - p2, v]).T
    b = np.array(p1)
    x = np.linalg.solve(A, b)

    # Debug:
    # try:
    #     x = np.linalg.solve(A, b)
    # except np.linalg.LinAlgError:
    #     raise ValueError
    # assert np.allclose((p2 - p1) * x[0] + p1, v * x[1])
    # assert np.allclose(np.dot(A, x), b)

    # vertices on the edge are possibly visible! ( < not <=)
    return x[1] < 1.0


def no_self_intersection(coords):
    polygon_length = len(coords)
    # again_check = []
    for index_p1, index_q1 in combinations(range(polygon_length), 2):
        # always: index_p1 < index_q1
        if index_p1 == index_q1 - 1 or index_p1 == index_q1 + 1:
            # neighbouring edges never have an intersection
            continue
        p1, p2 = coords[index_p1], coords[(index_p1 + 1) % polygon_length]
        q1, q2 = coords[index_q1], coords[(index_q1 + 1) % polygon_length]
        intersect_status = get_intersection_status(p1, p2, q1, q2)
        if intersect_status == 1:
            return False
        # if intersect_status == 2:
        # TODO 4 different options. check which side the other edge lies on.
        # if edge changes sides this is a an intersection
        # again_check.append((p1, p2, q1, q2))
        # print(p1, p2, q1, q2)

    # TODO check for intersections across 2 edges! use computed intersection

    return True


def has_clockwise_numbering(coords):
    """ tests if a polygon has clockwise vertex numbering

    approach: Sum over the edges, (x2 − x1)(y2 + y1). If the result is positive the curve is clockwise.
    from:
    https://stackoverflow.com/questions/1165647/how-to-determine-if-a-list-of-polygon-points-are-in-clockwise-order

    :param coords: the list of (x,y) coordinates representing the polygon to be tested
    :return: true if the polygon has been given in clockwise numbering
    """
    total_sum = 0.0
    p1 = coords[-1]
    for p2 in coords:
        x1, y1 = p1
        x2, y2 = p2
        total_sum += (x2 - x1) * (y2 + y1)
        p1 = p2
    return total_sum > 0


# TODO test
# todo - polygons must not intersect each other
def check_data_requirements(boundary_coords: np.ndarray, list_hole_coords: List[np.ndarray]):
    """ ensures that all the following conditions on the polygons are fulfilled:
        - must at least contain 3 vertices
        - no consequent vertices with identical coordinates in the polygons! In general might have the same coordinates
        - a polygon must not have self intersections (intersections with other polygons are allowed)
        - edge numbering has to follow this convention (for easier computations):
            * outer boundary polygon: counter clockwise
            * holes: clockwise
    :param boundary_coords:
    :param list_hole_coords:
    :return:
    """

    # TODO verbose error messages
    assert boundary_coords.shape[0] >= 3
    assert boundary_coords.shape[1] == 2
    assert no_identical_consequent_vertices(boundary_coords)
    assert no_self_intersection(boundary_coords)
    assert not has_clockwise_numbering(boundary_coords)

    for hole_coords in list_hole_coords:
        assert hole_coords.shape[0] >= 3
        assert hole_coords.shape[1] == 2
        assert no_identical_consequent_vertices(hole_coords)
        assert no_self_intersection(hole_coords)
        assert has_clockwise_numbering(hole_coords)

    # TODO rectification


def find_within_range(repr1, repr2, repr_diff, vertex_set, angle_range_less_180, equal_repr_allowed):
    """
    filters out all vertices whose representation lies within the range between
      the two given angle representations
    which range ('clockwise' or 'counter-clockwise') should be checked is determined by:
      - query angle (range) is < 180deg or not (>= 180deg)
    :param repr1:
    :param repr2:
    :param repr_diff: abs(repr1-repr2)
    :param vertex_set:
    :param angle_range_less_180: whether the angle between repr1 and repr2 is < 180 deg
    :param equal_repr_allowed: whether vertices with the same representation should also be returned
    :return:
    """

    if len(vertex_set) == 0:
        return vertex_set

    if repr_diff == 0.0:
        return set()

    min_repr_val = min(repr1, repr2)
    max_repr_val = max(repr1, repr2)  # = min_angle + angle_diff

    def lies_within(vertex):
        # vertices with the same representation will not NOT be returned!
        return min_repr_val < vertex.get_angle_representation() < max_repr_val

    def lies_within_eq(vertex):
        # vertices with the same representation will be returned!
        return min_repr_val <= vertex.get_angle_representation() <= max_repr_val

    # when the range contains the 0.0 value (transition from 3.99... -> 0.0)
    # it is easier to check if a representation does NOT lie within this range
    # -> filter_fct = not_within
    def not_within(vertex):
        # vertices with the same representation will NOT be returned!
        return not (min_repr_val <= vertex.get_angle_representation() <= max_repr_val)

    def not_within_eq(vertex):
        # vertices with the same representation will be returned!
        return not (min_repr_val < vertex.get_angle_representation() < max_repr_val)

    if equal_repr_allowed:
        lies_within_fct = lies_within_eq
        not_within_fct = not_within_eq
    else:
        lies_within_fct = lies_within
        not_within_fct = not_within

    if repr_diff < 2.0:
        # angle < 180 deg
        if angle_range_less_180:
            filter_fct = lies_within_fct
        else:
            # the actual range to search is from min_val to max_val, but clockwise!
            filter_fct = not_within_fct

    elif repr_diff == 2.0:
        # angle == 180deg
        # which range to filter is determined by the order of the points
        # since the polygons follow a numbering convention,
        # the 'left' side of p1-p2 always lies inside the map
        # -> filter out everything on the right side (='outside')
        if repr1 < repr2:
            filter_fct = lies_within_fct
        else:
            filter_fct = not_within_fct

    else:
        # angle > 180deg
        if angle_range_less_180:
            filter_fct = not_within_fct
        else:
            filter_fct = lies_within_fct

    return set(filter(filter_fct, vertex_set))


def convert_gridworld(size_x: int, size_y: int, obstacle_iter: iter, simplify: bool = True) -> (list, list):
    """
    prerequisites: grid world must not have non-obstacle cells which are surrounded by obstacles
    ("single white cell in black surrounding" = useless for path planning)
    :param size_x: the horizontal grid world size
    :param size_y: the vertical grid world size
    :param obstacle_iter: an iterable of coordinate pairs (x,y) representing blocked grid cells (obstacles)
    :param simplify: whether the polygons should be simplified or not. reduces edge amount, allow diagonal edges
    :return: an boundary polygon (counter clockwise numbering) and a list of hole polygons (clockwise numbering)
    NOTE: convert grid world into polygons in a way that coordinates coincide with grid!
        -> no conversion of obtained graphs needed!
        the origin of the polygon coordinate system is (-0.5,-0.5) in the grid cell system (= corners of the grid world)
    """

    assert size_x > 0 and size_y > 0

    if len(obstacle_iter) == 0:
        # there are no obstacles. return just the simple boundary rectangle
        return [np.array(x, y) for x, y in [(0, 0), (size_x, 0), (size_x, size_y), (0, size_y)]], []

    # convert (x,y) into np.arrays
    # obstacle_iter = [np.array(o) for o in obstacle_iter]
    obstacle_iter = np.array(obstacle_iter)

    def within_grid(pos):
        return 0 <= pos[0] < size_x and 0 <= pos[1] < size_y

    def is_equal(pos1, pos2):
        return np.all(pos1 == pos2)

    def pos_in_iter(pos, iter):
        for i in iter:
            if is_equal(pos, i):
                return True
        return False

    def is_obstacle(pos):
        return pos_in_iter(pos, obstacle_iter)

    def is_blocked(pos):
        return not within_grid(pos) or is_obstacle(pos)

    def is_unblocked(pos):
        return within_grid(pos) and not is_obstacle(pos)

    def find_start(start_pos, boundary_detect_fct, **kwargs):
        # returns the lowest and leftmost unblocked grid cell from the start position
        # for which the detection function evaluates to True
        start_x, start_y = start_pos
        for y in range(start_y, size_y):
            for x in range(start_x, size_x):
                pos = np.array([x, y])
                if boundary_detect_fct(pos, **kwargs):
                    return pos

    # north, east, south, west
    directions = np.array([[0, 1], [1, 0], [0, -1], [-1, 0]], dtype=int)
    # the correct offset to determine where nodes should be added.
    offsets = np.array([[0, 1], [1, 1], [1, 0], [0, 0]], dtype=int)

    def construct_polygon(start_pos, boundary_detect_fct, cntr_clockwise_wanted: bool):
        current_pos = start_pos.copy()
        # (at least) the west and south are blocked
        #   -> there has to be a polygon node at the current position (bottom left corner of the cell)
        edge_list = [start_pos]
        forward_index = 0  # start with moving north
        forward_vect = directions[forward_index]
        left_index = (forward_index - 1) % 4
        # left_vect = directions[(forward_index - 1) % 4]
        just_turned = True

        # follow the border between obstacles and free cells ("wall") until one reaches the start position again
        while 1:
            # left has to be checked first
            # do not check if just turned left or right (-> the left is blocked for sure)
            # left_pos = current_pos + left_vect
            if not (just_turned or boundary_detect_fct(current_pos + directions[left_index])):
                # print('< turn left')
                forward_index = left_index
                left_index = (forward_index - 1) % 4
                forward_vect = directions[forward_index]
                just_turned = True

                # add a new node at the correct position
                # decrease the index first!
                edge_list.append(current_pos + offsets[forward_index])

                # move forward (previously left, there is no obstacle)
                current_pos += forward_vect
            else:
                forward_pos = current_pos + forward_vect
                if boundary_detect_fct(forward_pos):
                    node_pos = current_pos + offsets[forward_index]
                    # there is a node at the bottom left corner of the start position (offset= (0,0) )
                    if is_equal(node_pos, start_pos):
                        # check and terminate if this node does already exist
                        break

                    # add a new node at the correct position
                    edge_list.append(node_pos)
                    # print('> turn right')
                    left_index = forward_index
                    forward_index = (forward_index + 1) % 4
                    forward_vect = directions[forward_index]
                    just_turned = True
                    # print(direction_index,forward_vect,just_turned,edge_list,)
                else:
                    # print('^ move forward')
                    current_pos += forward_vect
                    just_turned = False

        if cntr_clockwise_wanted:
            # make edge numbering counter clockwise!
            edge_list.reverse()
        return np.array(edge_list)

    # build the boundary polygon
    # start at the lowest and leftmost unblocked grid cell
    start_pos = find_start(start_pos=(0, 0), boundary_detect_fct=is_unblocked)
    # print(start_pos+directions[3])
    # raise ValueError
    boundary_edges = construct_polygon(start_pos, boundary_detect_fct=is_blocked, cntr_clockwise_wanted=True)

    if simplify:
        # TODO
        raise NotImplementedError()

    # detect which of the obstacles have to be converted into holes
    # just the obstacles inside the boundary polygon are part of holes
    # shift coordinates by +(0.5,0.5) for correct detection
    # the border value does not matter here
    unchecked_obstacles = [o for o in obstacle_iter if
                           inside_polygon(o[0] + 0.5, o[1] + 0.5, boundary_edges, border_value=True)]

    hole_list = []
    while len(unchecked_obstacles) > 0:
        start_pos = find_start(start_pos=(0, 0), boundary_detect_fct=pos_in_iter, iter=unchecked_obstacles)
        hole = construct_polygon(start_pos, boundary_detect_fct=is_unblocked, cntr_clockwise_wanted=False)

        # detect which of the obstacles still do not belong to any hole:
        # delete the obstacles which are included in the just constructed hole
        unchecked_obstacles = [o for o in unchecked_obstacles if
                               not inside_polygon(o[0] + 0.5, o[1] + 0.5, hole, border_value=True)]

        if simplify:
            # TODO
            pass

        hole_list.append(hole)

    return boundary_edges, hole_list


def find_visible(vertex_candidates, edges_to_check):
    """
    # IMPORTANT: self.translate(new_origin=query_vertex) always has to be called before!
        (for computing the angle representations wrt. the query vertex)
    query_vertex: a vertex for which the visibility to the vertices should be checked.
        also non extremity vertices, polygon vertices and vertices with the same coordinates are allowed.
        query point also might lie directly on an edge! (angle = 180deg)

    :param vertex_candidates: the set of all vertices which should be checked for visibility.
        IMPORTANT: is being manipulated, so has to be a copy!
        IMPORTANT: must not contain the query vertex!

    :param edges_to_check: the set of edges which determine visibility
    :return: a set of tuples of all vertices visible from the query vertex and the corresponding distance
    """

    visible_vertices = set()
    if len(vertex_candidates) == 0:
        return visible_vertices

    priority_edges = set()
    # goal: eliminating all vertices lying 'behind' any edge
    # TODO improvement in combination with priority: process edges roughly in sequence, but still allow jumps
    #  would follow closer edges more often which have a bigger chance to eliminate candidates -> speed up
    while len(vertex_candidates) > 0 and len(edges_to_check) > 0:
        # check prioritized items first
        try:
            edge = priority_edges.pop()
            edges_to_check.remove(edge)
        except KeyError:
            edge = edges_to_check.pop()

        lies_on_edge = False
        v1, v2 = edge.vertex1, edge.vertex2
        if v1.get_distance_to_origin() == 0.0:
            # vertex1 has the same coordinates as the query vertex -> on the edge
            lies_on_edge = True
            # (but does not belong to the same polygon, not identical!)
            # mark this vertex as not visible (would otherwise add 0 distance edge in the graph)
            vertex_candidates.discard(v1)
            # its angle representation is not defined (no line segment from vertex1 to query vertex!)
            range_less_180 = v1.is_extremity
            # do not check the other neighbouring edge of vertex1 in the future
            e1 = v1.edge1
            edges_to_check.discard(e1)
            priority_edges.discard(e1)
            # everything between its two neighbouring edges is not visible for sure
            v1, v2 = v1.get_neighbours()

        elif v2.get_distance_to_origin() == 0.0:
            lies_on_edge = True
            vertex_candidates.discard(v2)
            range_less_180 = v2.is_extremity
            e1 = v2.edge2
            edges_to_check.discard(e1)
            priority_edges.discard(e1)
            v1, v2 = v2.get_neighbours()

        repr1 = v1.get_angle_representation()
        repr2 = v2.get_angle_representation()

        repr_diff = abs(repr1 - repr2)
        if repr_diff == 2.0:
            # angle == 180deg -> on the edge
            lies_on_edge = True
            range_less_180 = False  # does actually not matter here

        if lies_on_edge:
            # when the query vertex lies on an edge (or vertex) no behind/in front checks must be performed!
            # the neighbouring edges are visible for sure
            try:
                vertex_candidates.remove(v1)
                visible_vertices.add(v1)
            except KeyError:
                pass
            try:
                vertex_candidates.remove(v2)
                visible_vertices.add(v2)
            except KeyError:
                pass

            # all the candidates between the two vertices v1 v2 are not visible for sure
            # candidates with the same representation should not be deleted, because they can be visible!
            vertex_candidates.difference_update(
                find_within_range(repr1, repr2, repr_diff, vertex_candidates, angle_range_less_180=range_less_180,
                                  equal_repr_allowed=False))
            continue

        # case: a 'regular' edge
        # eliminate all candidates which are blocked by the edge
        # that means inside the angle range spanned by the edge and actually behind it
        vertices_to_check = vertex_candidates.copy()
        # the vertices belonging to the edge itself (its vertices) must not be checked.
        # use discard() instead of remove() to not raise an error (they might not be candidates)
        vertices_to_check.discard(v1)
        vertices_to_check.discard(v2)
        if len(vertices_to_check) == 0:
            continue

        # assert repr1 is not None
        # assert repr2 is not None

        # for all candidate edges check if there are any candidate vertices (besides the ones belonging to the edge)
        #   within this angle range
        # the "view range" of an edge from a query point (spanned by the two vertices of the edge)
        #   is always < 180deg when the edge is not running through the query point (=180 deg)
        #  candidates with the same representation as v1 or v2 should be considered.
        #   they can be visible, but should be ruled out if they lie behind any edge!
        vertices_to_check = find_within_range(repr1, repr2, repr_diff, vertices_to_check, angle_range_less_180=True,
                                              equal_repr_allowed=True)
        if len(vertices_to_check) == 0:
            continue

        # if a candidate is farther away from the query point than both vertices of the edge,
        #    it surely lies behind the edge
        max_distance = max(v1.get_distance_to_origin(), v2.get_distance_to_origin())
        vertices_behind = set(filter(lambda extr: extr.get_distance_to_origin() > max_distance, vertices_to_check))
        # they do not have to be checked, no intersection computation necessary
        # TODO improvement: increase the neighbouring edges' priorities when there were extremities behind
        vertices_to_check.difference_update(vertices_behind)
        if len(vertices_to_check) == 0:
            # also done later, only needed if skipping this edge
            vertex_candidates.difference_update(vertices_behind)
            continue

        # if the candidate is closer than both edge vertices it surely lies in front (
        min_distance = min(v1.get_distance_to_origin(), v2.get_distance_to_origin())
        vertices_in_front = set(
            filter(lambda extr: extr.get_distance_to_origin() < min_distance, vertices_to_check))
        # they do not have to be checked (safes computation)
        vertices_to_check.difference_update(vertices_in_front)

        # for all remaining vertices v it has to be tested if the line segment from query point (=origin) to v
        #    has an intersection with the current edge p1---p2
        # vertices directly on the edge are allowed (not eliminated)!
        p1 = v1.get_coordinates_translated()
        p2 = v2.get_coordinates_translated()
        for vertex in vertices_to_check:
            if lies_behind(p1, p2, vertex.get_coordinates_translated()):
                vertices_behind.add(vertex)
            else:
                vertices_in_front.add(vertex)

        # vertices behind any edge are not visible
        vertex_candidates.difference_update(vertices_behind)
        # if there are no more candidates left. immediately quit checking edges
        if len(vertex_candidates) == 0:
            break

        # check the neighbouring edges of all vertices which lie in front of the edge next first
        # (prioritize them)
        # they lie in front and hence will eliminate other vertices faster
        # the fewer vertex candidates remain, the faster the procedure
        # TODO improvement: increase priority every time and draw highest priority items
        #   but this involves sorting (expensive for large polygons!)
        #   idea: work with a list of sets, add new set for higher priority, no real sorting, but still managing!
        # TODO test speed impact
        for e in vertices_in_front:
            # only add the neighbour edges to the priority set if they still have to be checked!
            if type(e) == PolygonVertex:
                # only vertices belonging to polygons have neighbours
                priority_edges.update(edges_to_check.intersection({e.edge1, e.edge2}))

    # all edges have been checked
    # all remaining vertices were not concealed behind any edge and hence are visible
    visible_vertices.update(vertex_candidates)

    # return a set of tuples: (vertex, distance)
    return {(e, e.get_distance_to_origin()) for e in visible_vertices}
